import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.util.ArrayList;

/**
 * A world filled with numbers to match together.
 * 
 * Author: Sarah Coakley 
 * Date: May 23, 2019
 */
public class NumberWorld extends World
{
    //Arrays used
    public ArrayList<Number> numbers = new ArrayList<Number>(0);
    private String[] keyName = {"g", "h", "j"};
    private String[] btnUpImage = {"check", "hint", "clear"};
    private String[] btnDownImage = {"check-down", "hint-down", "clear-down"};
    
    //Variables used
    BtnLight light = new BtnLight();
    Highlight hi = new Highlight();
    Highlight hi2 = new Highlight();
    private int actualWidth;
    private int actualHeight;
    
    public int cellSize = 45;
    private int topSpace = 38;
    private int activeNumber = -1;
    
    private boolean gPressed = false;
    private boolean hPressed = false;
    private boolean jPressed = false;

    public NumberWorld()
    {    
        // Create a new world with 399x578 cells with a cell size of 1x1 pixels.
        super(399, 668, 1);
        
        setPaintOrder(Number.class, Highlight.class);
        
        spawnButtons();
        initializeNumbers();
    }

    private void spawnButtons()
    {
        //Loop to spawn in a check, hint, and clear button
        for(int i = 0; i < keyName.length; i++)
        {
            Button btn = new Button(keyName[i], btnUpImage[i] + ".png", btnDownImage[i] + ".png");
            
            //Gets the height and width of the button's image and spawns it in accordingly
            actualWidth = btn.getActualWidth();
            actualHeight = btn.getActualHeight();
            addObject(btn, 66 + i * actualWidth, actualHeight - 19);
        }
    }
    
    private void initializeNumbers()
    {
        //Adds numbers to the ArrayList
        for(int i = 1; i < 19; i++)
        {
            //If 10 is encountered, skip it
            if(i == 10) 
            {
                continue;
            }
            
            //Splits the numbers above 10 into separate digits
            int[] j = {i/10, i%10};
            for(int k = 0; k < j.length; k++)
            {
                //Makes the image display a number without a 0 in front of it
                if(j[k] == 0) 
                {
                    continue;
                }
                
                numbers.add(new Number(j[k]));
            }
        }
        
        //Spawns in the numbers from the ArrayList to the world
        for(int i = 0; i < numbers.size(); i++)
        {
            addNumber(i);
        }
    }
    
    public void act()
    {
        //If the g key is pressed, add numbers and check if the puzzle is still possible
        if(gPressed != Greenfoot.isKeyDown("g"))
        {
            gPressed = !gPressed;
            
            //Add numbers to the world
            if(gPressed)
            {
                addNumbers();
                
                if(!matchPossible())
                {
                    Greenfoot.stop();
                }
            }
            
            //If the check button is highlighted and pressed, remove the highlight
            if(light != null)
            {
                removeObject(light);
            }
        }
        
        //If the h key is pressed, hint to a match or the check button
        if(hPressed != Greenfoot.isKeyDown("h"))
        {
            hPressed = !hPressed;
            
            if(hPressed)
            {
                //If there is a possible match, highlight it
                if(matchPossible())
                {
                    hiMatch();
                }
                
                //If the are no matches, highlight the check button
                if(!matchPossible())
                {
                    addObject(light, 66, 19);
                }
            }
        }
        
        //If the j key is pressed, remove any solved numbers
        if(jPressed != Greenfoot.isKeyDown("j"))
        {
            jPressed = !jPressed;
            
            if(jPressed)
            {
                removeSolved();
            }
        }
        
        //If all the numbers are solved, stop the program
        if(allSolved())
        {
            Greenfoot.stop();
        }
    }
    
    public void pressedNumber(Number number)
    {
        //If there is no active number, keep the image unhighlighted
        if(activeNumber == -1)
        {
            changeActive(number);
        }
        
        //If the first number is clicked again, reest the image
        else if(number == numbers.get(activeNumber))
        {
            changeActive(null);
        }
        
        //If two clicked numbers don't match, reest the image of any highlighted number
        else if(number.getValue() != numbers.get(activeNumber).getValue()
            && number.getValue() + numbers.get(activeNumber).getValue() != 10)
        {
            changeActive(number);
        }
        
        //Checks if two clicked numbers are a match and can be removed or not
        else
        {
            int[] n = {activeNumber, numbers.indexOf(number)};
            boolean match = false;
            
            //Checks the columns for matches
            if(n[0]%9 == n[1]%9)
            {
                //Makes the match
                match = true;
                
                //Prevents a possible match is there is an unsolved number in between them
                for(int i = n[0]/9; i != n[1]/9; i = i+(int)Math.signum(n[1]-n[0]))
                {
                    //If there is no unsolved number, make the match
                    if(i == n[0]/9) 
                    {
                        continue;
                    }
                    
                    //Cancel the match if there is an unsolved number in between matchable numbers
                    if(!numbers.get(i*9+n[0]%9).getSolved())
                    { //unsolved number in between
                        changeActive(number);
                        match = false;
                        break;
                    }
                }
            }
            
            //Checks the rows for matches if there are no matches in the columns
            if(!match)
            {
                //Make the match
                match = true;
                
                //Prevents a possible match is there is an unsolved number in between them
                for(int i = n[0]; i != n[1]; i = i+(int)Math.signum(n[1]-n[0]))
                {
                    //If there is no unsolved number, make the match
                    if(i == n[0]) 
                    {
                        continue;
                    }
                    
                    //Cancel the match if there is an unsolved number in between matchable numbers
                    if(!numbers.get(i).getSolved())
                    {
                        changeActive(number);
                        match = false;
                        break;
                    }
                }
            }
            
            //If there is no unsolved number between two numbers, enable the ability to match them
            if(match)
            {
                solve(number);
                
                //If there is a highlight, remove it
                if(hi != null)
                {
                    removeObject(hi);
                }
                
                //If there is a highlight, remove it
                if(hi2 != null)
                {
                    removeObject(hi2);
                }
            }
        }
    }
    
    public boolean matchPossible()
    {
        //For loop to check if there are any matches left
        for(int i = 0; i < numbers.size(); i++)
        {
            //Makes sure already solved numbers can't be matched again
            if(numbers.get(i).getSolved()) 
            {
                continue;
            }
            
            //Checks the numbers horizontally for matches
            for(int j = i+1; j < numbers.size(); j++)
            {
                //Ignore all solved numbers
                if(numbers.get(j).getSolved()) 
                {
                    continue;
                }
                
                //If there is a match found, set matchPossible to true
                if(numbers.get(i).getValue() == numbers.get(j).getValue()
                    || numbers.get(i).getValue() + numbers.get(j).getValue() == 10)
                { 
                    return true;
                }
                
                //If there isn't a match, break the loop
                else 
                {
                    break;
                }
            }
            
            //Checks the numbers vertically for matches
            for(int j = i+9; j < numbers.size(); j+=9)
            {
                //Ignores any solved numbers
                if(numbers.get(j).getSolved()) 
                {
                    continue;
                }
                
                //If there is a match found, set matchPossible to true
                if(numbers.get(i).getValue() == numbers.get(j).getValue()
                    || numbers.get(i).getValue() + numbers.get(j).getValue() == 10)
                {
                    return true;
                }
                
                //If there isn't a match, break the loop
                else 
                {
                    break;
                }
            }
        }
        
        //Returns false if there are no available matches
        return false;
    }
    
    public boolean allSolved()
    {
        //For loop to check if there are any matches left
        for(int i = 0; i < numbers.size(); i++)
        {
            //Makes sure already solved numbers can't be matched again
            if(numbers.get(i).getSolved()) 
            {
                continue;
            }
            
            //Checks the numbers horizontally for matches
            for(int j = i+1; j < numbers.size(); j++)
            {
                //Ignore all solved numbers
                if(numbers.get(j).getSolved()) 
                {
                    continue;
                }
                
                //If there is a match found, set allSolved to false
                if(numbers.get(i).getValue() != 0 && numbers.get(j).getValue() != 0)
                { 
                    return false;
                }
                
                //If there isn't a match, break the loop
                else 
                {
                    break;
                }
            }
            
            //Checks the numbers vertically for matches
            for(int j = i+9; j < numbers.size(); j+=9)
            {
                //Ignores any solved numbers
                if(numbers.get(j).getSolved()) 
                {
                    continue;
                }
                
                //If there is a match found, set allSolved to false
                if(numbers.get(i).getValue() != 0 && numbers.get(j).getValue() != 0)
                {
                    return false;
                }
                
                //If there isn't a match, break the loop
                else 
                {
                    break;
                }
            }
        }
        
        //Returns true if there are no unsolved numbers
        return true;
    }
    
    public void hiMatch()
    {
        for(int i = 0; i < numbers.size(); i++)
        {
            //Makes sure already solved numbers can't be matched again
            if(numbers.get(i).getSolved()) 
            {
                continue;
            }
            
            //Checks the numbers horizontally for matches
            for(int j = i+1; j < numbers.size(); j++)
            {
                //Ignore all solved numbers
                if(numbers.get(j).getSolved()) 
                {
                    continue;
                }
                
                //If there is a match found, set a highlight on a match
                if(numbers.get(i).getValue() == numbers.get(j).getValue()
                    || numbers.get(i).getValue() + numbers.get(j).getValue() == 10)
                { 
                    //Get the coordinates of the first matchable number and add a highlight
                    int ix = numbers.get(i).getX();
                    int iy = numbers.get(i).getY();
                    addObject(hi, ix, iy);
                    
                    //Get the coordinates of the second matchable number and add a highlight
                    int jx = numbers.get(j).getX();
                    int jy = numbers.get(j).getY();
                    addObject(hi2, jx, jy);
                }
                
                //If there isn't a match, break the loop
                else 
                {
                    break;
                }
            }
            
            //Checks the numbers vertically for matches
            for(int j = i+9; j < numbers.size(); j+=9)
            {
                //Ignores any solved numbers
                if(numbers.get(j).getSolved()) 
                {
                    continue;
                }
                
                //If there is a match found, set a highlight on a match
                if(numbers.get(i).getValue() == numbers.get(j).getValue()
                    || numbers.get(i).getValue() + numbers.get(j).getValue() == 10)
                {
                    //Get the coordinates of the first matchable number and add a highlight
                    int ix = numbers.get(i).getX();
                    int iy = numbers.get(i).getY();
                    addObject(hi, ix, iy);
                    
                    //Get the coordinates of the second matchable number and add a highlight
                    int jx = numbers.get(j).getX();
                    int jy = numbers.get(j).getY();
                    addObject(hi2, jx, jy);
                }
                
                //If there isn't a match, break the loop
                else 
                {
                    break;
                }
            }
        }
    }
    
    private void addNumber(int i)
    {
        //Spawns in a number from the ArrayList to the world in the correct place
        addObject(numbers.get(i), i%9*(cellSize) + cellSize/2, topSpace + i/9*(cellSize) + cellSize/2);
    }
    
    private void addNumbers()
    {
        int size = numbers.size();
        
        //Adds all unsolved numbers to the ArrayList
        for(int i = 0; i < size; i++)
        {
            if(!numbers.get(i).getSolved()) 
            {
                numbers.add(new Number(numbers.get(i).getValue()));
            }
        }
        
        //Spawns in all the numbers from the ArrayList
        for(int i = size; i < numbers.size(); i++)
        {
            addNumber(i);
        }
    }
    
    private void solve(Number second)
    {
        //Covers both numbers and puts activeNumber back to -1
        numbers.get(activeNumber).setSolved();
        second.setSolved();
        activeNumber = -1;
    }
    
    public void removeSolved()
    {
        //Loop to remove solved numbers
        for(int i = numbers.size()-1; i > -1; i--)
        {
            //Ignores unsolved numbers
            if(!numbers.get(i).getSolved()) 
            {
                continue;
            }
            
            //Removes the solved number from the world and the ArrayList
            else
            {
                numbers.remove(numbers.get(i));
                removeObject(numbers.get(i));
            }
        }
        
        //Removes all unsolved numbers
        removeObjects(getObjects(Number.class));
        
        //Re-adds the removed numbers in the proper order
        for(int i = 0; i < numbers.size(); i++)
        {
            addNumber(i);
        }
    }
    
    private void changeActive(Number newActive)
    {
        //If a number is not active, keep it inactive
        if(activeNumber != -1) 
        {
            numbers.get(activeNumber).updateImage(false);
        }
        
        //If there is a new active number, change its image
        if(newActive != null) 
        {
            newActive.updateImage(true);
        }
        
        activeNumber = newActive != null ? numbers.indexOf(newActive) : -1;
    }
}
